<?php
// qtkrparse.php version 3.0.1
// for documentation etc. see http://www.erik-krause.de/qtkrparse/

/* LICENSE:

Copyright (c) 2008-2016 Erik Krause
Portions Copyright (c) 2005 Aldo Hoeben

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The origin of this software must not be misrepresented; you must not claim
that you wrote the original software. If you use this software in a product,
an acknowledgment in the product documentation would be appreciated but is
not required.

Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.


THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.

*/

/*
CONTRIBUTORS:

Erik Krause
Klaus Reinfeld
Aldo Hoeben
Thomas Rauscher
Ingemar Bergmark
Joost Nieuwenhuijse
*/

//XML start
$xml = "<krpano onstart='cm_save_variables'>\n";
$UsePreview = true;

$inc = @fopen('krpano.inc', 'rb');
if ($inc) {
  $inc_xml = fread($inc, filesize('krpano.inc'));
  $inc_time = filemtime('krpano.inc');
  fclose($inc);
  if ($inc_xml) $xml .= $inc_xml;
}

//These <view> attributes are used:
$fisheye = "0"; //default="0.35
$fisheyefovlink = "3"; //default="0.5"
$limitfov = "true";
$camroll="0.0";

//error_reporting(E_ALL);
$old_error_handler = set_error_handler("myErrorHandler");

$UsePreview == ($UsePreview && ini_get("allow_url_fopen") && (imagetypes() & IMG_JPG));

// Get quicktime file name
$file = (array_key_exists("mov", $_GET)) ? $_GET["mov"] : "";
// basename() also strips \x00, we don't need to worry about ? and # in path:
// Must be real files anyway, fopen() does not support wildcards

$a = explode('.', basename($file));
$ext = array_pop($a);

//myErrorHandler (-1,$ext);

$path=$_SERVER['REQUEST_URI'];
$path=substr($path, 0, strpos($path, '/qtkrparse.php'));
chdir($_SERVER['DOCUMENT_ROOT'].$path);

if (strcasecmp($ext, "mov") != 0 || !file_exists($file))
{
  myErrorHandler (-1,"File not found or no .mov extension: " . $file,"","");
}

// Open quicktime file for parsing
$fp = fopen($file, "rb");


// Parse the file and output XML or get jpeg tile from file?
$action = (array_key_exists("action", $_GET)) ? $_GET["action"] : "";
$ofsfile = substr($file, 0, strlen($file)-(strlen($ext)+1))."_{$ext}_kr.ofs";
if ($action!="tile") {
  $xmlfile = substr($file, 0, strlen($file)-(strlen($ext)+1))."_{$ext}_kr.xml";
  $PrvFile = substr($file, 0, strlen($file)-(strlen($ext)+1))."_{$ext}_prv.jpg";

  if (file_exists($PrvFile) && $UsePreview) {
    $xml .= "<preview type='cubestrip' url='$PrvFile' details='10' striporder='frblud'/>\n";
  }

  $cache = (array_key_exists("cache", $_GET)) ? $_GET["cache"] : "true";

  // cache=clear: just remove cache without creating a new file
  if ($cache=="clear") {
    if (file_exists($xmlfile)) {
      unlink($xmlfile);
      myErrorHandler(-1,"Cache removed.","","");
    } else
      myErrorHandler(-1,"Cache not found.","","");
  }

  if (file_exists($xmlfile) && file_exists($ofsfile) && ($cache=="true")) {
    // get oldest created file time (xml or offset)
    $CreatedFileTime = min(filemtime($xmlfile), filemtime($ofsfile));
    // get youngest user changed file
    $ChangedFileTime = max(getlastmod(), filemtime($file), ($inc)?$inc_time:getlastmod());
    // If cached files are older than user changed files, reset cache
    if ($CreatedFileTime < $ChangedFileTime)
      $cache = "reset";
  }

  if ($UsePreview) {
    if (file_exists($PrvFile) && ($cache != "reset")) {
      // If Preview file does not exist, is older than mov file, or older than qtkrparse.php file, reset cache
      if ($cache=="true" && ((filemtime($PrvFile) < filemtime($file)) || (filemtime($PrvFile) < getlastmod())) )
        $cache="reset";
    } else $cache = "uncertain"; //if a preview track will be found, cache will be invalidated.
  }
  if ( (!file_exists($xmlfile)) || (!file_exists($ofsfile)) || ($cache!="true") ) {
    // Parse quicktime file
    $fileindex = 0;
    do {
      fseek($fp, $fileindex);
      $atom = fread($fp,8);
      $atom = unpack("N1size/N1type", $atom);
      $atom["type"] = pack("N", $atom["type"]);

      $fileindex += $atom["size"];
    } while ($atom["type"] != "moov");

    if ($atom["type"] != "moov")
      myErrorHandler("No 'moov' atom found. Not a Quicktime movie?");

    $moov = fread($fp, $atom["size"]-8);

    $cmovA = GetAtom($moov, "cmov", 0);
    if($cmovA[0] != "")
      myErrorHandler("Compressed movie headers are not supported.");

    $panotrak = "";
    $videtrak = "";
    $pvidetrak = "";
    $trakindex = 0;

    $traks = array();
    $traknr = 0;

    $videtraknr = -1;
    $pvidetraknr = -1;
    $hotttraknr = -1;

    do {
      // Traverse all "trak" atoms.
      $trakA = GetAtom($moov, "trak", $trakindex);
      $trak = $trakA[0];
      $trakindex = $trakA[1];

      if ($trak != "") {
        $traknr += 1;

        // See what type of trak this is
        $hdlrA = GetAtom($trak, "hdlr", 0);
        $hdlr = $hdlrA[0];

        if ($hdlr != "") {
          $hdlr = unpack("N1/N1type/N1sub", $hdlr);
          $hdlr["type"] = pack("N", $hdlr["type"]);
          $hdlr["sub"] = pack("N", $hdlr["sub"]);

          if ($hdlr["type"]=="mhlr") {
            $type = $hdlr["sub"];
            if ($type=="pano") {
              $panotrak = $trak[0];

              $stcoA = GetAtom($trak, "stco", 0);
              $stco = unpack("N2/N1offset",$stcoA[0]);
              $pdatoffset = $stco["offset"];

              $stszA = GetAtom($trak, "stsz", 0);
              $stsz = unpack("N1/N1size",$stszA[0]);
              $pdatsize = $stsz["size"];

              $imgtA = GetAtom($trak, "imgt", 0);
              $imgt = unpack("N*", $imgtA[0]);
              if(count($imgt)>2)
                list(,$videtraknr,$pvidetraknr) = $imgt;
              else
                list(,$videtraknr) = $imgt;

              $hottA = GetAtom($trak, "hott", 0);
              if ($hottA[0] != "") {
                $hott = unpack("N*", $hottA[0]);
                list(,$hotttraknr) = $hott;
              } //if ($hottA[0] != "")
            } //if ($type=="pano")
          } //if ($hdlr["type"]=="mhlr")
        } //if ($hdlr != "")

        $traks[$traknr] = $trak;

        if ($videtraknr > -1 && $traknr >= $videtraknr && $traknr >= $pvidetraknr) {
          $videtrak = $traks[$videtraknr];
          if($pvidetraknr > -1)
            $pvidetrak = $traks[$pvidetraknr];
          break;
        }
      }

      $trakindex+=4;
    } while ($trak != "");

    // Got the tracks we're looking for?
    if ($panotrak == "" || $videtrak == "")
      myErrorHandler(-1,"Could not find 'pano' and/or 'vide' trak atoms. Not a Quicktime VR panorama?");

    // Is the video track we got in JPEG format?
    $stsdA = GetAtom($videtrak, "stsd", 0);
    $stsd = unpack("N3/N1format/N6/n1width/n1height", $stsdA[0]);
    $stsd["format"] = pack("N" , $stsd["format"]);
    if ($stsd["format"] != "jpeg")
      myErrorHandler(-1,"Only jpeg compressed panoramas supported.");

    // Read panorama description track
    fseek($fp, $pdatoffset);
    $panodat = fread($fp, $pdatsize);

    $pdatA = GetAtom($panodat, "pdat", 0);
    $pdat = $pdatA[0];

    if ($pdat == "")
      myErrorHandler(-1,"Could not find 'pdat' atom. Not a Quicktime VR panorama?");

    $pdat = FlipFloats(unpack("N3/n2ver/N2ref/f9cam/N2imgsize/n2imgframes/N2hotsize/n2hotframes/N1flags/N1type", $pdat));
    $pdat["type"] = pack("N", $pdat["type"]);

    if ($pdat["type"] != "cube")
      myErrorHandler(-1,"Not a cubic panorama. Cylinders are not yet supported.");


    $cuvwA = GetAtom($panodat, "cuvw", $pdat[1]);
    $cuvw = $cuvwA[0];

    if ($cuvw == "")
      myErrorHandler(-1,"Could not find 'cuvw' atom. Invalid cubic panorama.");

    $cuvw = FlipFloats(unpack("N3/f9cam",$cuvw));

    $cufaA = GetAtom($panodat, "cufa", $pdat[1]);
    $cufa = $cufaA[0];

    $tiles = array();
    if ($cufa == "") {
      // Could not find cufa atom. Set 'standard' cubic face data.
      $tiles[] = array("orientation1" => 1 , "orientation2" => 0 , "orientation3" => 0 , "orientation4" => 0, "center1" => 0, "center2" => 0);
      $tiles[] = array("orientation1" => 0.5, "orientation2" => 0 , "orientation3" =>-0.5, "orientation4" => 0, "center1" => 0, "center2" => 0);
      $tiles[] = array("orientation1" => 0 , "orientation2" => 0 , "orientation3" => 1 , "orientation4" => 0, "center1" => 0, "center2" => 0);
      $tiles[] = array("orientation1" => 0.5, "orientation2" => 0 , "orientation3" => 0.5, "orientation4" => 0, "center1" => 0, "center2" => 0);
      $tiles[] = array("orientation1" => 0.5, "orientation2" => 0.5, "orientation3" => 0 , "orientation4" => 0, "center1" => 0, "center2" => 0);
      $tiles[] = array("orientation1" => 0.5, "orientation2" =>-0.5, "orientation3" => 0 , "orientation4" => 0, "center1" => 0, "center2" => 0);
    } else {
      // Get face data from cufa.
      for ($index = 12; $index <= (strlen($cufa)-12); $index+=32) {
        $tiles[] = FlipFloats(unpack("f4orientation/f2center/f1aspect/f1skew", substr($cufa, $index, 32)));
      }
    }

    if (count($tiles) == 0)
      myErrorHandler(-1,"No tiles found in cubic panorama.");

    // Fix invalid values for in pdat atom (generated by Pano2QTVR 0.9.5.3 and older)
    if ((($pdat["imgframes1"] / 4) * $pdat["imgframes2"] * 6) != count($tiles)) {
      $pdat["imgframes1"] = sqrt(count($tiles) / 6) * 4;
      $pdat["imgframes2"] = $pdat["imgframes1"] / 4;
    }


    // Build xml string from Quicktime data
    $stcoA = GetAtom($videtrak, "stco", 0);
    $stszA = GetAtom($videtrak, "stsz", 0);
    $stscA = GetAtom($videtrak, "stsc", 0);

    $stco_num = unpack("N1/N1num", $stcoA[0]);
    $stco_num = $stco_num["num"];
    $stsz_num = unpack("N2/N1num", $stszA[0]);
    $stsz_num = $stsz_num["num"];
    $stsc_num = unpack("N1/N1num", $stscA[0]);
    $stsc_num = $stsc_num["num"];

    $stco = unpack("N2/N".$stco_num."tile", $stcoA[0]);
    $stsz = unpack("N3/N".$stsz_num."tile", $stszA[0]);
    $stsc = unpack("N2/N".($stsc_num * 3)."s2c", $stscA[0]);


    if ($stsz_num != count($tiles))
      myErrorHandler(-1,"Inconsistent number of tiles in tracks.");

    if ($stco_num != $stsz_num) {
      $stsc["s2c" . (1 + $stsc_num * 3)] = $stco_num+1;

      $stsz_index=0;
      $stco_index=0;
      $new_stco = array();

      for($s2c = 0; $s2c < $stsc_num; $s2c++) {
        for($chunk=$stsc["s2c" . (1 + $s2c * 3)]; $chunk<$stsc["s2c" . (4 + $s2c * 3)]; $chunk++) {
          $stco_index++;
          $base = $stco["tile".$stco_index];
          //$base=0;
          $offset=0;

          $c_numtiles = $stsc["s2c" . (2 + $s2c * 3)];
          for($tile = 0; $tile < $c_numtiles; $tile++)
          {
            $stsz_index++;
            $new_stco["tile".$stsz_index] = $base + $offset;
            $offset += $stsz["tile".$stsz_index];
          }
        }
      }

      $stco = $new_stco;
    }


    if (($pvidetrak != "") && $UsePreview) {
      $pstcoA = GetAtom($pvidetrak, "stco", 0);
      $pstszA = GetAtom($pvidetrak, "stsz", 0);
      $pstscA = GetAtom($pvidetrak, "stsc", 0);

      $pstco_num = unpack("N1/N1num", $pstcoA[0]);
      $pstco_num = $pstco_num["num"];
      $pstsz_num = unpack("N2/N1num", $pstszA[0]);
      $pstsz_num = $pstsz_num["num"];
      $pstsc_num = unpack("N1/N1num", $pstscA[0]);
      $pstsc_num = $pstsc_num["num"];


      $pstco = unpack("N2/N".$pstco_num."tile", $pstcoA[0]);
      $pstsz = unpack("N3/N".$pstsz_num."tile", $pstszA[0]);
      $pstsc = unpack("N2/N".($pstsc_num * 3)."s2c", $pstscA[0]);

      if ($pstco_num != $pstsz_num) {
        $pstsc["s2c" . (1 + $pstsc_num * 3)] = $pstco_num+1;

        $pstsz_index=0;
        $pstco_index=0;
        $new_pstco = array();

        for($s2c = 0; $s2c < $pstsc_num; $s2c++) {
          for($chunk=$pstsc["s2c" . (1 + $s2c * 3)]; $chunk<$pstsc["s2c" . (4 + $s2c * 3)]; $chunk++) {
            $pstco_index++;
            $base = $pstco["tile".$pstco_index];
            //$base=0;
            $offset=0;

            $c_numtiles = $pstsc["s2c" . (2 + $s2c * 3)];
            for($tile = 0; $tile < $c_numtiles; $tile++)
            {
              $pstsz_index++;
              $new_pstco["tile".$pstsz_index] = $base + $offset;
              $offset += $pstsz["tile".$pstsz_index];
            }
          }
        }

        $pstco = $new_pstco;
      }
      if ($cache='uncertain') $cache = "reset";
    } else // if (($pvidetrak != "") && $UsePreview)
      if ($cache='uncertain') $cache = "true";

    $JpegOfs = array("front" => array(),
                     "right" => array(),
                     "back"  => array(),
                     "left"  => array(),
                     "up"    => array(),
                     "down"  => array()
                    );

    $PrvOfs = $JpegOfs;

    $hlookatmin = round($cuvw["cam1"],4);
    $hlookatmax = round($cuvw["cam2"],4);
    $vlookatmin = round($cuvw["cam3"],4);
    $vlookatmax = round($cuvw["cam4"],4);
    $limitview = ((($hlookatmax - $hlookatmin) < 360) || (($vlookatmax - $vlookatmin) < 180))?"range":"lookat";

    $xml.= "\t<view "
        ."hlookat='".-round($cuvw["cam7"],4)."' "
        ."vlookat='".-round($cuvw["cam8"],4)."' "
        ."camroll='$camroll' "
        ."fov='".round($cuvw["cam9"],4)."' "
        ."fovmin='".round($cuvw["cam5"],4)."' "
        ."fovmax='".round($cuvw["cam6"],4)."' "
        ."fisheye='$fisheye' "
        ."fisheyefovlink='$fisheyefovlink' "
        ."limitfov='$limitfov' "
        ."limitview='$limitview' "
        ."hlookatmin='$hlookatmin' "
        ."hlookatmax='$hlookatmax' "
        ."vlookatmin='$vlookatmin' "
        ."vlookatmax='$vlookatmax' "
        ."/>\n"
        ."\t<description>Converted from ".$file." by qtkrparse.php</description>\n"
        ."\t<!-- some variables for later use -->\n"
        ."\t<action name='cm_save_variables'>\n"
        ."\t\t set(origfov, ".round($cuvw["cam9"],4).");\n"
        ."\t\t set(origfovmin, ".round($cuvw["cam5"],4).");\n"
        ."\t\t set(origfovmax, ".round($cuvw["cam6"],4).");\n"
        ."\t</action>\n";

    $hfov = 2 * rad2deg(atan(1 / ($pdat["imgframes1"]/4)));
    $vfov = 2 * rad2deg(atan(1 / $pdat["imgframes2"]));
    $baseindex = 100;

    $TilesPerCubeSide = sqrt(count($tiles)/6);

    // 1, if even amount of tiles, 0 if odd
    $EvenTiles = (($TilesPerCubeSide+1) % 2);

    for ($panoelem = 0; $panoelem < count($tiles); $panoelem++) {
      // For each tile, add panoelement with inline image node

      $W = $tiles[$panoelem]["orientation1"];
      $X = $tiles[$panoelem]["orientation2"];
      $Y = $tiles[$panoelem]["orientation3"];
      $Z = $tiles[$panoelem]["orientation4"];

      if  ($W != 0 && $X == 0 && $Y == 0) {
//        $pan = "0"; $tilt = "0";
        $face = "front";
      } else if ($W > 0 && $X == 0 && $Y ==-$W) {
//        $pan = "90"; $tilt = "0";
        $face = "right";
      } else if ($W == 0 && $X == 0 && $Y != 0) {
//        $pan = "180"; $tilt = "0";
        $face = "back";
      } else if ($W != 0 && $X == 0 && $Y == $W) {
//        $pan = "-90"; $tilt = "0";
        $face = "left";
      } else if ($W != 0 && $X == $W && $Y == 0) {
//        $pan = "0"; $tilt = "90";
        $face = "up";
      } else if ($W != 0 && $X ==-$W && $Y == 0) {
//        $pan = "0"; $tilt = "-90";
        $face = "down";
      }

      // 4 subtiles are numbered -3 -1 1 3
      // 5 subtiles are numbered -4 -2 0 2 4
      // we need consecutive numbers
      $x_ofs = ($tiles[$panoelem]["center1"] + $EvenTiles) / 2;
      $y_ofs = (-$tiles[$panoelem]["center2"] + $EvenTiles) / 2;

      if ($baseindex > $x_ofs) $baseindex = $x_ofs;

      // Tile offsets and lengths
      $JpegOfs[$face][$x_ofs][$y_ofs]["start"] = $stco["tile".($panoelem+1)];
      $JpegOfs[$face][$x_ofs][$y_ofs]["length"] = $stsz["tile".($panoelem+1)];
      // Preview offsets and lenghts
      if (($pvidetrak!="") && $UsePreview) {
        $PrvOfs[$face][$x_ofs][$y_ofs]["pstart"] = $pstco["tile".($panoelem+1)];
        $PrvOfs[$face][$x_ofs][$y_ofs]["plength"] = $pstsz["tile".($panoelem+1)];
      }
/**/
    } //for ($panoelem = 0; $panoelem < count($tiles); $panoelem++)

    $tiledimagesize = count($JpegOfs["left"]) * $stsd["width"];
    $xml.= "\t<image type='CUBE' multires='true' baseindex='$baseindex' tilesize='".$stsd["width"]."' tiledimagewidth='$tiledimagesize' tiledimageheight='$tiledimagesize'>\n"
          ."\t<level tiledimagewidth='$tiledimagesize' tiledimageheight='$tiledimagesize'>\n"
          ."\t\t<left  url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=left&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t\t<front url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=front&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t\t<right url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=right&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t\t<back  url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=back&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t\t<up    url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=up&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t\t<down  url='qtkrparse.php/tile.jpg?mov=$file&amp;action=tile&amp;face=down&amp;x_ofs=%x&amp;y_ofs=%y' />\n"
          ."\t</level>\n"
          ."\t</image>\n"
          ."</krpano>";

    if($cache=="true" || $cache=="reset") {
      $filepointer = fopen($ofsfile, "wb");
      fwrite($filepointer, serialize($JpegOfs));
      fclose($filepointer);
      //if Preview don't write immediately in order to force new creation with preview tag
      if (file_exists($PrvFile) || !$UsePreview) {
        // Cache XML file for next visits
        $filepointer = fopen($xmlfile, "wb");
        fwrite($filepointer, $xml);
        fclose($filepointer);
      }
      //if
      if (($pvidetrak != "") && $UsePreview) {
        ini_set('memory_limit', '50M');
        //prepare URL to fetch preview subtiles.
        $httppath = "http://".$_SERVER['HTTP_HOST'].$path."/qtkrparse.php/tile.jpeg?mov=".$file."&action=tile";
        //get one tile to determine size
        $imgsrc = $httppath."&start=".$PrvOfs["left"][0][0]["pstart"];
        $imgsrc.= "&length=".$PrvOfs["left"][0][0]["plength"];
        $SubPrv = imagecreatefromjpeg($imgsrc);
        $SubTileSize = imagesx($SubPrv);
        imagedestroy($SubPrv);
        //calculate tile size
        $PrvSize = $SubTileSize * (count($PrvOfs["left"][0]));
        $PrvImg = imagecreatetruecolor(6 * $PrvSize, $PrvSize);
        //loop all faces
        $FacePos = 0;
        foreach ($PrvOfs as $face => $facedata) {
          $XPos = 0;
          foreach ($facedata as $PrvY => $tiles) {
            $YPos = 0;
            foreach ($tiles as $PrvY => $data) {
              //create URL
              $imgsrc = $httppath."&start=".$data["pstart"];
              $imgsrc.= "&length=".$data["plength"];
              //get image
              $SubPrv = imagecreatefromjpeg($imgsrc);
              //copy images in place
              imagecopy($PrvImg, $SubPrv, $FacePos+$XPos, $YPos, 0, 0, $SubTileSize, $SubTileSize);
              imagedestroy($SubPrv);
              $YPos += $SubTileSize;
            }
            $XPos += $SubTileSize;
          }
          $FacePos += $PrvSize;
        }

/*
        foreach ($PrvOfs as $PrvTile) {
          //loop rows per tile
          $YPos = 0;
          for ($PrvY = $baseindex; $PrvY <= ($baseindex * -1); $PrvY++) {
            //loop columns per tile
            $XPos = 0;
            for ($PrvX = $baseindex; $PrvX <= ($baseindex * -1); $PrvX++) {
              //create URL
              $imgsrc = $httppath."&start=".$PrvTile[$PrvX][$PrvY]["pstart"];
              $imgsrc.= "&length=".$PrvTile[$PrvX][$PrvY]["plength"];
              //get image
              $SubPrv = imagecreatefromjpeg($imgsrc);
              //copy images in place
              imagecopy($PrvImg, $SubPrv, $FacePos+$XPos, $YPos, 0, 0, $SubTileSize, $SubTileSize);
              imagedestroy($SubPrv);
              $XPos += $SubTileSize;
            }
            $YPos += $SubTileSize;
          }
          $FacePos += $PrvSize;
        } // foreach ($PrvOfs as $PrvTile)
*/
        imagejpeg($PrvImg, $PrvFile, 20);
      } //if (($pvidetrak != "") && $UsePreview)
    }

    // Output XML file
    header("Content-type: text/xml");
    header("Content-Length: ".strlen($xml));
    header("Content-Disposition: inline; filename=$xmlfile");
    echo $xml;

  } else {
    // If cached XML file is available, supply that instead of parsing
//    header("Location: ".$xmlfile);
    header("Content-type: text/xml");
    header("Content-Length: ".strlen($xml));
    header("Content-Disposition: inline; filename=$xmlfile");
    echo file_get_contents($xmlfile); /**/
    exit();
  }

} else {
  restore_error_handler();
  $i=0;
  do {
    $filepointer = @fopen($ofsfile, "rb");
    $i++;
  } while (!$filepointer && ($i < 100));
  if ($filepointer) {
    $JpegOfs = unserialize(fread($filepointer, filesize($ofsfile)));
    fclose($filepointer);

    $face = (array_key_exists("face", $_GET)) ? $_GET["face"] : "left";
    $x_ofs = (array_key_exists("x_ofs", $_GET)) ? $_GET["x_ofs"] : "0";
    $y_ofs = (array_key_exists("y_ofs", $_GET)) ? $_GET["y_ofs"] : "0";


    $jpgstart = (array_key_exists("start", $_GET)) ? $_GET["start"] : $JpegOfs[$face][$x_ofs][$y_ofs]["start"];
    $jpglength = (array_key_exists("length", $_GET)) ? $_GET["length"] : $JpegOfs[$face][$x_ofs][$y_ofs]["length"];

    $CacheFileTime = max(filemtime($file), filemtime($ofsfile));

    // Getting headers sent by the client.
    if (function_exists('apache_request_headers'))
      $headers = apache_request_headers();

    // Checking if the client is validating his cache and if it is current.
    if ($headers && isset($headers['If-Modified-Since']) && (strtotime($headers['If-Modified-Since']) == $CacheFileTime)) {
        // Client's cache IS current, so we just respond '304 Not Modified'.
        header('Last-Modified: '.gmdate('D, d M Y H:i:s', $CacheFileTime).' GMT', true, 304);
    } else {
      // Image not cached or cache outdated, we respond '200 OK' and output the image.
      header('Last-Modified: '.gmdate('D, d M Y H:i:s', $CacheFileTime).' GMT', true, 200);
      header("Expires: Mon, 24 Jul 2017 05:00:00 GMT"); // Date in the future
      header("Accept-Ranges: bytes");
    header("Content-Type: image/jpeg");
    header("Content-Length: $jpglength");
    header("Content-Disposition: inline; filename=tile_{$x_ofs}_{$y_ofs}.jpg");

    fseek($fp, $jpgstart);
    echo fread($fp, $jpglength);
    }

  } else {
    $im = @imagecreatetruecolor(250, 250) or die("Cannot Initialize new GD image stream");
    $text_color = imagecolorallocate($im, 255, 255, 255);
    imagestring($im, 5, 5, 5,  "Cannot open", $text_color);
    imagestring($im, 5, 5, 20,  "$ofsfile", $text_color);
    header("Content-Type: image/jpeg");
    header("Content-Disposition: inline");
    imagejpeg($im);
  }
}

fclose($fp);

function myErrorHandler($errno, $errstr, $errfile="", $errline="")
{
  $ErrTypes = array(E_WARNING           => "Warning",
                    E_NOTICE            => "Notice",
                    E_USER_ERROR        => "User_Error",
                    E_USER_WARNING      => "User_Warning",
                    E_USER_NOTICE       => "User_Notice");
  if (($errno & ini_get('error_reporting')) > 0) //internal, respect error_reporting setting
    $errstr = $ErrTypes[$errno].": ".strip_tags($errstr)." in $errfile on line $errline";
  else if ($errno < 0) //direct call
    $errstr = "Error: ," . $errstr;
  else return true;

  $srch = array(',',     "'",'"');
  $repl = array( '',"&apos;","&quot;");
  $errstr = str_replace($srch,$repl,$errstr);

  // Output XML file
  header("Content-type: text/xml");
  header("Content-Disposition: inline");
  echo "<krpano onstart=\"error($errstr)\">";
  echo "</krpano>";
  exit ();
}

function GetAtom($qt, $atomType, $offset) {
  $atomIndex = strpos($qt, $atomType, $offset);

  if ($atomIndex != FALSE) {
    $atomSize = unpack("N*",substr($qt, $atomIndex - 4, 4));
    $atomSize = $atomSize[1]-4;
    if ($atomSize != 1)
      return array( substr($qt, $atomIndex + 4, $atomSize) , $atomIndex);
  } else {
    return array("",0);
  }
}

function FlipFloats($array) {
  list ($endiantest) = array_values (unpack ('L1L', pack ('V',1)));
  if ($endiantest == 1) {
    for ($item = 0; $item < count($array); $item++) {
      $values = array_values($array);
      $keys = array_keys($array);

      if (is_float($values[$item])) {
        $repacked = unpack("f*", strrev(pack("f", $values[$item])));
        $array[$keys[$item]] = $repacked[1];
      }
    }
  }
  return $array;
}

?>
